import time

from sanic.exceptions import NotFound, InvalidUsage

from prometheus_sanic.constants import RequestMetaData, BaseMetrics


def before_request_handler(request):
    request[RequestMetaData.TIME] = time.time()

    # Fetch handler from router
    try:
        _handler, _args, _kwargs, uri = request.app.router.get(request)
        request[RequestMetaData.ROUTER] = uri
    except NotFound:
        request[RequestMetaData.ROUTER] = None
    except InvalidUsage:
        request[RequestMetaData.ROUTER] = None


def after_request_endpoint_handler(request, response, get_endpoint_fn,
                                   *args, **kwargs):
    lat = time.time() - request[RequestMetaData.TIME]
    endpoint = get_endpoint_fn(request)

    # Note, that some handlers can ignore response logic,
    # for example, websocket handler
    response_status = response.status if response else 200
    request.app.metrics[BaseMetrics.LATENCY.name].labels(
        request.method, endpoint, response_status
    ).observe(lat)
    request.app.metrics[BaseMetrics.COUNT.name].labels(
        request.method, endpoint, response_status
    ).inc()


def after_request_uri_handler(request, response, *args, **kwargs):
    lat = time.time() - request[RequestMetaData.TIME]

    uri = request[RequestMetaData.ROUTER]

    # Note, that some handlers can ignore response logic,
    # for example, websocket handler
    response_status = response.status if response else 200
    request.app.metrics[BaseMetrics.LATENCY.name].labels(
        request.method, uri, response_status
    ).observe(lat)
    request.app.metrics[BaseMetrics.COUNT.name].labels(
        request.method, uri, response_status
    ).inc()



from enum import Enum


class RequestMetaData:
    TIME = '__START_TIME__'
    ROUTER = '__ROUTER_URI__'


class BaseMetrics(Enum):
    LATENCY = 'sanic_request_latency_sec'
    COUNT = 'sanic_request_count'



import os

from prometheus_client import CONTENT_TYPE_LATEST, CollectorRegistry, \
    core, multiprocess, start_http_server
from prometheus_client.exposition import generate_latest
from sanic.request import Request
from sanic.response import raw

from . import endpoint, metrics, handlers
from .constants import BaseMetrics
from .exceptions import SanicPrometheusError


class MonitorSetup:
    def __init__(self, app, metrics_path, multiprocess_on=False):
        self._app = app
        self._multiprocess_on = multiprocess_on
        self._metrics_path = metrics_path

    def expose_endpoint(self):
        """
        Expose /metrics endpoint on the same Sanic server.

        This may be useful if Sanic is launched from a container
        and you do not want to expose more than one port for some
        reason.
        """

        @self._app.route(self._metrics_path, methods=['GET'])
        async def expose_metrics(request):
            return raw(self._get_metrics_data(),
                       content_type=CONTENT_TYPE_LATEST)

    def start_server(self, addr='', port=8000):
        """
        Expose /metrics endpoint on a new server that will
        be launched on `<addr>:<port>`.

        This may be useful if you want to restrict access to
        metrics data with firewall rules.
        NOTE: can not be used in multiprocessing mode
        """
        if self._multiprocess_on:
            raise SanicPrometheusError(
                "start_server can not be used when "
                "multiprocessing is turned on"
            )
        start_http_server(addr=addr, port=port)

    def _get_metrics_data(self):
        if not self._multiprocess_on:
            registry = core.REGISTRY
        else:
            registry = CollectorRegistry()
            multiprocess.MultiProcessCollector(registry)
        data = generate_latest(registry)
        return data


def monitor(app, endpoint_type='url:1',
            get_endpoint_fn=None,
            latency_buckets=None,
            multiprocess_mode='all',
            metrics_path='/metrics',
            is_middleware=True,
            metrics_list=None,
            before_handler=None,
            after_handler=None,
            base_metrics=None,
            init_metrics_object: dict = None
            ):
    """
    Regiesters a bunch of metrics for Sanic server
    (request latency, count, etc) and exposes /metrics endpoint
    to allow Prometheus to scrape them out.

    :param init_metrics_object:
    :param base_metrics:
    :param before_handler:
    :param after_handler:
    :param metrics_list:
    :param is_middleware:
    :param metrics_path:
    :param multiprocess_mode:
    :param app: an instance of sanic.app
    :param endpoint_type: All request related metrics have a label called
                         'endpoint'. It can be fetched from Sanic `request`
                          object using different strategies specified by
                          `endpoint_type`:
                            url - full relative path of a request URL
                                (i.e. for http://something/a/b/c?f=x you end up
                                having `/a/b/c` as the endpoint)
                            url:n - like URL but with at most `n` path elements
                                in the endpoint (i.e. with `url:1`
                                http://something/a/b/c becomes `/a`).
                            custom - custom endpoint fetching funciton that
                                should be specified by `get_endpoint_fn`
    :param get_endpoint_fn: a custom endpoint fetching function that is ignored
                            until `endpoint_type='custom'`.
                              get_endpoint_fn = lambda r: ...
                            where `r` is Sanic request object
    :param latency_buckets: an optional list of bucket sizes for latency
                            histogram (see prometheus `Histogram` metric)

    :multiprocess_mode':
    :metrics_path:         path where metrics will be exposed

    NOTE: memory usage is not collected when when multiprocessing is enabled
    """
    multiprocess_on = 'prometheus_multiproc_dir' in os.environ
    get_endpoint = endpoint.fn_by_type(endpoint_type, get_endpoint_fn)

    if base_metrics is None:
        base_metrics = BaseMetrics

    if before_handler is None:
        before_handler = handlers.before_request_handler

    if after_handler is None:
        after_handler = handlers.after_request_endpoint_handler

    @app.listener('before_server_start')
    def before_start(init_app, _loop):
        init_app.metrics = {}
        if init_metrics_object is not None:
            init_app.metrics = init_metrics_object
        metrics.init(
            init_app, latency_buckets, multiprocess_mode,
            metrics_list, base_metrics
        )

    if is_middleware is True:
        @app.middleware('request')
        async def before_request(request: Request):
            if request.path != metrics_path and request.method != "OPTIONS":
                before_handler(request)

        @app.middleware('response')
        async def before_response(request: Request, response):
            if request.path != metrics_path and request.method != "OPTIONS":
                after_handler(request, response, get_endpoint)

    if multiprocess_on:
        @app.listener('after_server_stop')
        def after_stop(_app, _loop):
            multiprocess.mark_process_dead(os.getpid())

    return MonitorSetup(app, metrics_path, multiprocess_on)


__all__ = ['monitor']



def get_from_url(request, lim=0):
    if lim <= 0:
        return request.path
    url = request.path
    if url[0] == '/':
        lim += 1
    return '/'.join(url.split('/')[:lim])


def make_url_endpoint_fn(url_type):
    lim_str = url_type[4:]
    if not lim_str:
        return get_from_url
    else:
        try:
            n = int(lim_str)
        except ValueError:
            raise ValueError(
                'url endpoint type can be either "url" or "url:<int>"'
            )
        return lambda req: get_from_url(req, lim=n)


def fn_by_type(ep_type, get_endpoint_fn):
    if ep_type.startswith('url'):
        return make_url_endpoint_fn(ep_type)
    elif ep_type == 'custom':
        if get_endpoint_fn is None:
            raise ValueError(
                '"custom" type requires endpoint function to be set'
            )
        return get_endpoint_fn
    else:
        raise ValueError('Unsupported endpoint type "{}"'.format(ep_type))


__all__ = ['fn_by_type', 'get_from_url']



from typing import List, Tuple

from prometheus_client import Counter, Histogram
from prometheus_client.metrics import MetricWrapperBase


def init(app, latency_buckets=None, multiprocess_mode='all',
         metrics_list: List[Tuple[str, MetricWrapperBase]] = None,
         metrics=None):
    app.metrics[metrics.COUNT.name] = Counter(
        metrics.COUNT.value,
        'Sanic Request Count',
        ['method', 'endpoint', 'http_status']
    )

    hist_kwargs = {}
    if latency_buckets is not None:
        hist_kwargs = {'buckets': latency_buckets}
    app.metrics[metrics.LATENCY.name] = Histogram(
        metrics.LATENCY.value,
        'Sanic Request Latency Histogram',
        ['method', 'endpoint', 'http_status'],
        **hist_kwargs
    )

    if metrics_list:
        for name, pm_metric in metrics_list:
            # todo set multiprocess_mode is Gauge
            # app.metrics[name'] = Gauge(
            #             ...
            #             multiprocess_mode=multiprocess_mode
            #         )
            app.metrics[name] = pm_metric



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prometheus_sanic']

package_data = \
{'': ['*']}

install_requires = \
['prometheus-client>=0.7.1,<0.8.0', 'sanic>=0.7.0']

setup_kwargs = {
    'name': 'prometheus-sanic',
    'version': '2.1.1',
    'description': 'Exposes Prometheus monitoring metrics of Sanic apps.',
    'long_description': 'Sanic prometheus metrics\n=========================\n|Build Status| |PyPI| |PyPI version|\n\nAfter googling for a while I didn\'t find a library that would enable some `prometheus <https://prometheus.io/>`_ metrics for `Sanic <https://github.com/channelcat/sanic>`_-based apps, so I had to write one. It makes adding monitoring to your Sanic app super easy, just add one line to your code (ok, two if you count import :) and point Prometheus to a newly appeared `/metrics` endpoint.\n\nInstallation and enabling monitoring\n-------------------------------------\n\nInstallatio:\n\n.. code:: bash\n\n  pip install prometheus-sanic\n\nEasy-peasy:\n\n.. code:: python\n\n  from sanic import Sanic\n  from prometheus_sanic import monitor\n\n  app = Sanic()\n  ...\n\n  if __name__ == "__main__":\n    monitor(app).expose_endpoint() # adds /metrics endpoint to your Sanic server\n    app.run(host="0.0.0.0", port=8000)\n\n\nActually, there\'re two ways to run monitoring:\n\n\n1. The one you\'ve seen above, ``monitor(app).expose_endpoint()``. \n   It just adds a new ``route`` to your Sanic app, exposing ``/metrics`` endpoint\n   on the same host and port your Sanic server runs. It might be useful if you run your\n   app in a container and you do not want to expose different ports for metrics and everything else.\n   You can customize the ``/metrics`` endpoint by passing the ``metrics_path`` keyword argument:\n   ``monitor(app, metrics_path=\'/my_metrics_path\').expose_endpoint()``.\n2. ``monitor(app).start_server(addr=..., port=...)``.\n   Runs a HTTP server on given address and port and exposes ``/metrics`` endpoint on it.\n   This might be useful if you want to restrict access to your ``/metrics`` endpoint using some\n   firewall rules\n\n\nVersions compatibility\n----------------------\n\n* ☑︎ use **0.1.0** for Sanic <= 0.4.1\n* ☑︎ use **0.1.3** for Sanic >= 0.5.0\n* ☑︎ use >= **0.1.4** if you need multiprocessing support\n* ☑︎ use **0.1.6** if you have to use `promtheus-client` <= 0.4.2\n* ☑︎ use **0.1.8** with `prometheus-client` >= 0.5.0\n* ☑︎ use **0.2.0** with `prometheus-client` >= 0.7.1 and Sanic >= 18.12\n\nExposed metrics\n-----------------\n\nAt the moment ``prometheus-sanic`` provides four metrics:\n\n* **sanic_request_count** - total number of requests (labels: *method*, *endpoint*, *status*) [`counter <https://prometheus.io/docs/concepts/metric_types/#counter>`_]\n* **sanic_request_latency_sec** - request latency in seconds (labels: *method*, *endpoint*) [`histogram <https://prometheus.io/docs/concepts/metric_types/#histogram>`_]\n* **sanic_mem_rss_bytes** - resident memory used by the process (in bytes) [`gauge <https://prometheus.io/docs/concepts/metric_types/#gauge>`_]\n* **sanic_mem_rss_perc** - a percent of total physical memory used by the process running Sanic [`gauge <https://prometheus.io/docs/concepts/metric_types/#gauge>`_]\n  \nLabels\n-----------------\n\n* **method**: a HTTP method (i.e. GET/POST/DELETE/etc)\n* **endpoint**: just a string, a name identifying a point handling a group of requests. By default it\'s just the first element of the relative path of the URL being called (i.e. for http://myhost/a/b/c you\'ll end up having ``/a`` as your endpoint). It is quite configurable, in fact it\'s up you what\'s gonna get to the ``endpoint`` label (see ``help(prometheus_sanic.monitor)`` for more details)\n* **http_status**: a HTTP status code\n\nMultiprocess mode\n-----------------\n\nSanic allows to launch multiple worker processes to utilise parallelisation, which is great but makes metrics collection much trickier (`read more <https://github.com/prometheus/client_python/blob/master/README.md#multiprocess-mode-gunicorn>`_) and introduces some limitations. \n\nIn order to collect metrics from multiple workers, create a directory and point a ``prometheus_multiproc_dir`` environment variable to it. Make sure the directory is empty before you launch your service::\n\n\n     % rm -rf /path/to/your/directory/*\n     % env prometheus_multiproc_dir=/path/to/your/directory python your_sanic_app.py\n\n\nUnfortunately you can not use ``monitor(app).start_server(addr=..., port=...)`` in multiprocess mode as it exposes a prometheus endpoint from a newly created process.\n\nConfiguration\n-----------------\n\nBest you can do is::\n\n     % ipython\n     In [1]: from prometheus_sanic import monitor\n     In [2]: help(monitor)\n\n\nPrometheus quering examples:\n-----------------------------\n\n* *Average latency over last 30 minutes*::\n\n    rate(sanic_request_latency_sec_sum{endpoint=\'/your-endpoint\'}[30m]) / \n    rate(sanic_request_latency_sec_count{endpoint=\'/your-endpoint\'}[30m])\n\n* *95th percentile of request latency*:: \n\n    histogram_quantile(0.95, sum(rate(sanic_request_latency_sec_bucket[5m])) by (le))\n\n* *Physical memory usage percent over last 10 minutes*::\n\n    rate(sanic_mem_rss_perc[10m])\n\n.. |Build Status| image:: https://github.com/skar404/prometheus-sanic/workflows/Tests/badge.svg\n   :target: https://github.com/skar404/prometheus-sanic/actions/\n.. |PyPI| image:: https://img.shields.io/pypi/v/prometheus-sanic.svg\n   :target: https://pypi.python.org/pypi/prometheus-sanic/\n.. |PyPI version| image:: https://img.shields.io/pypi/pyversions/prometheus-sanic.svg\n   :target: https://pypi.python.org/pypi/prometheus-sanic/\n',
    'author': 'Dan Kruchinin',
    'author_email': 'dan.kruchinin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/skar404/prometheus-sanic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
